package net.cattaka.rdbassistant.script.core;

import java.awt.Color;
import java.util.List;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import net.cattaka.jspf.JspfException;
import net.cattaka.rdbassistant.gui.table.ResultSetTableModel;
import net.cattaka.swing.table.StdStyledCell;
import net.cattaka.swing.util.KeyStringListComparator;
import net.cattaka.util.MessageBundle;

/**
 * ScriptTableに関する汎用的な処理を提供します。
 * 
 * @author cattaka
 */
public class ScriptTableUtil {
	/** デフォルトの等しい場合に使用するマークです */
	private CellMark equalCellMark;
	/** デフォルトの等しくない場合に使用するマークです */
	private CellMark notEqualCellMark;
	/** デフォルトの追加分に使用するマークです */
	private CellMark addedCellMark;
	/** デフォルトの削除分に使用するマークです */
	private CellMark deletedCellmark;
	/** デフォルトの枠線付きセルに使用するマークです */
	private CellMark borderCellMark;
	
	public ScriptTableUtil() {
		equalCellMark = new CellMark();
		notEqualCellMark = new CellMark();
		addedCellMark = new CellMark();
		deletedCellmark = new CellMark();
		borderCellMark = new CellMark();
		notEqualCellMark.setBackground(Color.YELLOW);
		addedCellMark.setBackground(Color.GREEN);
		deletedCellmark.setBackground(Color.RED);
		borderCellMark.setBorderThickness(1);
	}
	
	/**
	 * 内部処理用にResultSetTableModelをScriptTableに変換する。
	 * 
	 * @param rs
	 * @return
	 */
	ScriptTable createScriptTable(ResultSetTableModel rs) {
		ScriptTable result = new ScriptTable(rs.getRowCount()+1, rs.getColumnCount()+1);
		result.setValueAt("", 0, 0);
		// 列ヘッダ
		for (int c=1;c<=rs.getColumnCount();c++) {
			result.setValueAt(rs.getColumnName(c-1), 0, c);
		}

		// 行ヘッダ
		for (int r=1;r<=rs.getRowCount();r++) {
			result.setValueAt(String.valueOf(r), r, 0);
		}
		
		// テーブルの内容
		for (int r=0;r<rs.getRowCount();r++) {
			for (int c=0;c<rs.getColumnCount();c++) {
				Object obj = rs.getValueAt(r, c);
				if (obj != null) {
					result.setValueAt(obj.toString(), r+1, c+1);
				}
			}
		}
		// ボーダーの設定
		for (int r=0;r<result.getRowCount();r++) {
			for (int c=0;c<result.getColumnCount();c++) {
				StdStyledCell cell = result.getCellAt(r,c); 
				borderCellMark.mark(cell);
				result.setCellAt(cell,r,c);
			}
		}
		return result;
	}

	/**
	 * 指定されたテーブル間の差異を探し、差異のある箇所に指定したマーキングを行ったテーブルを返します。
	 * 行ヘッダと列ヘッダは一つ目のテーブルの物が使用されます。
	 * 削除された列も表示されます。
	 * 
	 * @param tbl1 比較対象となるテーブルです 
	 * @param tbl2 比較対象となるテーブルです 
	 * @param keyColumns キー項目の列名
	 * @return 比較結果
	 * @throws JspfException 指定された列名の列が存在しない場合やキー項目がユニークでない場合にスローされます。
	 */
	public ScriptTable createMarkedDiff(ScriptTable tbl1, ScriptTable tbl2, String... keyColumns) throws JspfException {
		int[] keyColumnIndices = new int[keyColumns.length];
		for (int i=0;i<keyColumns.length;i++) {
			keyColumnIndices[i] = tbl1.getColumnIndex(keyColumns[i]);
		}
		return createMarkedDiff(tbl1, tbl2, keyColumnIndices);
	}

	/**
	 * 指定されたテーブル間の差異を探し、差異のある箇所に指定したマーキングを行ったテーブルを返します。
	 * 行ヘッダと列ヘッダは一つ目のテーブルの物が使用されます。
	 * 削除された列も表示されます。
	 * 
	 * @param tbl1 比較対象となるテーブルです 
	 * @param tbl2 比較対象となるテーブルです 
	 * @param keyColumns キー項目の列番
	 * @return 比較結果
	 * @throws JspfException キー項目がユニークでない場合にスローされます。
	 */
	public ScriptTable createMarkedDiff(ScriptTable tbl1, ScriptTable tbl2, int... keyColumnIndices) throws JspfException {
		if (keyColumnIndices.length == 0) {
			keyColumnIndices = new int[]{0};
		}
		// テーブル１のユニークキーチェック
		{
			TreeSet<List<String>> uniqueTbl1 = new TreeSet<List<String>>(new KeyStringListComparator(keyColumnIndices));
			for (int r=1;r<tbl1.getRowCount();r++) {
				List<String> cols = tbl1.getRowValuesAsFullList(r);
				if (!uniqueTbl1.add(cols)) {
					throw new JspfException(MessageBundle.getMessage("keys_are_not_unique"));
				}
			}
		}
		// テーブル２のマップを作成
		TreeMap<List<String>, List<String>> indexedMap = new TreeMap<List<String>, List<String>>(new KeyStringListComparator(keyColumnIndices));
		for (int r=1;r<tbl2.getRowCount();r++) {
			List<String> cols = tbl2.getRowValuesAsFullList(r);
			if (indexedMap.put(cols,cols) != null) {
				throw new JspfException(MessageBundle.getMessage("keys_are_not_unique"));
			}
		}
		
		ScriptTable result = new ScriptTable(tbl1.getRowCount()+tbl2.getRowCount(), tbl1.getColumnCount());
	
		// 列のコピー
		for (int c=0;c<tbl2.getColumnCount();c++) {
			StdStyledCell cell = tbl1.getCellAt(0,c); 
			borderCellMark.mark(cell);
			result.setCellAt(cell,0,c);
		}
		
		// 内容のコピーと比較
		int rowCount = 0;
		for (int r=1;r<tbl1.getRowCount();r++) {
			rowCount++;
			
			List<String> cols1 = tbl1.getRowValuesAsFullList(r);
			List<String> cols2 = indexedMap.remove(cols1);
			StdStyledCell cell = result.getCellAt(r,0);
			cell.setValue(String.valueOf(rowCount));
			borderCellMark.mark(cell);
			result.setCellAt(cell,r,0);
			if (cols2 != null) {
				// 既存の列
				for (int c=1;c<cols1.size();c++) {
					String str1 = cols1.get(c);
					StdStyledCell cell2 = result.getCellAt(r,c);
					borderCellMark.mark(cell2);
					cell2.setValue(str1);
					if (c<cols2.size()) {
						String str2 = cols2.get(c);
						if (str1 == null && str2 == null) {
							equalCellMark.mark(cell2);
						} else if ((str1 != null && str2 == null) || (str1 == null && str2 != null)) {
							notEqualCellMark.mark(cell2);
						} else if (str1.equals(str2)) {
							equalCellMark.mark(cell2);
						} else {
							notEqualCellMark.mark(cell2);
						}
					} else {
						notEqualCellMark.mark(cell2);
					}
					result.setCellAt(cell2,r,c);
				}
			} else {
				// 追加された列
				for (int c=0;c<cols1.size();c++) {
					String str1 = cols1.get(c);
					StdStyledCell cell2 = result.getCellAt(r,c);
					cell2.setValue(str1);
					addedCellMark.mark(cell2);
					borderCellMark.mark(cell2);
					result.setCellAt(cell2, r, c);
				}
			}
		}
		// 削除された列
		Set<List<String>> keys = indexedMap.keySet();
		for (List<String>key : keys) {
			rowCount++;
			List<String> cols2 = indexedMap.get(key);
			result.setValueAt("", rowCount, 0);
			StdStyledCell cell = result.getCellAt(rowCount, 0);
			deletedCellmark.mark(cell);
			borderCellMark.mark(cell);
			result.setCellAt(cell, rowCount, 0);
			for (int c=1;c<cols2.size();c++) {
				String str1 = cols2.get(c);
				StdStyledCell cell2 = result.getCellAt(rowCount, c);
				cell2.setValue(str1);
				deletedCellmark.mark(cell2);
				borderCellMark.mark(cell2);
				result.setCellAt(cell2, rowCount, c);
			}
		}
		rowCount++;
		result.setSize(rowCount, result.getColumnCount());
		
		return result;
	}
}
